__all__ = [
    "multimodal_vectorizer"
]

from typing import Any, Set, List, Type, Optional

from radient.tasks.vectorizers._base import Vectorizer
from radient.tasks.vectorizers.audio.imagebind import ImageBindAudioVectorizer
from radient.tasks.vectorizers.image.imagebind import ImageBindImageVectorizer
from radient.tasks.vectorizers.text.imagebind import ImageBindTextVectorizer

IMAGEBIND_VECTORIZERS = {
    ImageBindAudioVectorizer,
    ImageBindImageVectorizer,
    ImageBindTextVectorizer
}


class MultimodalVectorizer(Vectorizer):

    def __init__(self, vectypes: Set[Type], **kwargs):
        super().__init__()
        self._vectorizers = {}
        for VecType in vectypes:
            vectorizer = VecType(**kwargs)
            self._vectorizers[vectorizer.vtype] = vectorizer

    def modalities(self) -> List[str]:
        return list(self._vectorizers.keys())

    def _preprocess(self, data: Any, modality: str, **kwargs) -> Any:
        vectorizer = self._vectorizers[modality]
        return vectorizer._preprocess(data, **kwargs)

    def _vectorize(self, data: Any, modality: str, **kwargs) -> Any:
        vectorizer = self._vectorizers[modality]
        vector = vectorizer._vectorize(data, **kwargs)
        return vector

    def _postprocess(self, data: Any, modality: str, **kwargs) -> Any:
        vectorizer = self._vectorizers[modality]
        return vectorizer._postprocess(data, **kwargs)

    def accelerate(self, **kwargs):
        for vectorizer in self._vectorizers.values():
            vectorizer.accelerate(**kwargs)


def multimodal_vectorizer(method: Optional[str], **kwargs) -> MultimodalVectorizer:
    """Creates a text vectorizer specified by `method`.
    """

    if method in ("imagebind",):
        return MultimodalVectorizer(IMAGEBIND_VECTORIZERS, **kwargs)
    else:
        raise NotImplementedError


from typing import Dict, Optional, Tuple, Type 

from radient.orchestrate.runners import *
from radient.tasks.sinks import *
from radient.tasks.sources import *
from radient.tasks.transforms import *
from radient.tasks.vectorizers import *



def make_operator(
    task_name: str,
    task_type: str,
    runner: Optional[Type] = None,
    task_params: Optional[Dict] = None
) -> Runner:

    runner = runner or LocalRunner
    task_params = task_params or {}

    # Create a data sink.
    if task_name == "sink":
        if task_type == "milvus":
            return runner(MilvusSink, task_params=task_params)
        else:
            raise ValueError(f"unknown data store: {task_type}")

    # Create a data source.
    elif task_name == "source":
        if task_type == "local":
            return runner(LocalSource, task_params=task_params)
        elif task_type == "youtube":
            return runner(YoutubeSource, task_params=task_params)
        else:
            raise ValueError(f"unknown data source: {task_type}")

    # Create a data-to-data transformation.
    elif task_name == "transform":
        if task_type == "video-demux":
            return runner(video_demux_transform, task_params=task_params)
        else:
            raise ValueError(f"unknown transform method: {task_type}")

    # Create an data-to-vector transformation.
    elif task_name == "vectorizer":
        if task_type == "audio":
            return runner(audio_vectorizer, task_params=task_params)
        elif task_type == "graph":
            return runner(graph_vectorizer, task_params=task_params)
        elif task_type == "image":
            return runner(image_vectorizer, task_params=task_params)
        elif task_type == "molecule":
            return runner(molecule_vectorizer, task_params=task_params)
        elif task_type == "text":
            return runner(text_vectorizer, task_params=task_params)
        elif task_type == "multimodal":
            return runner(multimodal_vectorizer, task_params=task_params)
        else:
            raise NotImplementedError




__all__ = [
    "TimmImageVectorizer"
]

from typing import Any, List

from radient.tasks.accelerate import export_to_onnx, ONNXForward
from radient.tasks.vectorizers.image._base import ImageVectorizer
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector

Image = LazyImport("PIL", attribute="Image", package_name="Pillow")
timm = LazyImport("timm")
torch = LazyImport("torch")


class TimmImageVectorizer(ImageVectorizer):
    """Computes image embeddings using `timm`.
    """

    def __init__(self, model_name: str = "resnet50", **kwargs):
        super().__init__()
        self._model_name = model_name
        self._model = timm.create_model(model_name, pretrained=True, **kwargs)
        self._model.reset_classifier(0)
        self._model.eval()
        data_config = timm.data.resolve_model_data_config(self._model)
        self._transform = timm.data.create_transform(**data_config)

    def _vectorize(self, image: Image, **kwargs) -> Vector:
        # TODO(fzliu): dynamic batching
        with torch.inference_mode():
            x = self._transform(image.convert("RGB")).unsqueeze(0)
            vector = self._model(x).squeeze()
            if isinstance(vector, torch.Tensor):
                vector = vector.numpy()
        return vector.view(Vector)

    def accelerate(self):
        # `timm` models take a single 4D tensor (`B x C x H x W`) as input.
        onnx_model_path = export_to_onnx(
            self,
            torch.randn((1, 3, 224, 224)),
            axes_names=["batch_size"],
            input_names=["images"],
            output_names=["vectors"],
            model_type="pytorch"
        )

        self._model.forward = ONNXForward(
            onnx_model_path
        )




__all__ = [
    "VoyageTextVectorizer"
]

from typing import List

from radient.tasks.vectorizers.text._base import TextVectorizer
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector

voyageai = LazyImport("voyageai")


class VoyageTextVectorizer(TextVectorizer):
    """Text vectorization with Voyage AI (https://www.voyageai.com).
    """

    def __init__(self, model_name: str = "voyage-2", **kwargs):
        super().__init__()
        self._model_name = model_name
        self._client = voyageai.Client()

    def _vectorize(self, text: str, **kwargs) -> Vector:
        res = self._client.embed(text, model=self._model_name)
        return np.array(res.embeddings).view(Vector)

    @property
    def model_name(self):
        return self._model_name



from collections.abc import Iterator
from pathlib import Path
from typing import Dict
import uuid

from radient.tasks.sources._base import Source
from radient.utils.lazy_import import LazyImport

yt_dlp = LazyImport("yt_dlp")


class YoutubeSource(Source):
    """Downloads videos from Youtube to a local directory. The `url` argument
    can be a single video or a playlist.
    """

    def __init__(self,
        url: str,
        output_directory: str = "~/.radient/data/youtube",
        **kwargs
    ):
        super().__init__()

        # Create a new output directory for downloading and storing the videos.
        output_directory = Path(output_directory).expanduser()
        output_directory = output_directory / str(uuid.uuid4())
        output_directory.mkdir(parents=True, exist_ok=True)

        # The input URL may be a single video or a playlist of videos. Here, we
        # extract a list of all video URLs for use in the `read` function.
        with yt_dlp.YoutubeDL({"extract_flat": True, "quiet": True}) as ydl:
            info = ydl.extract_info(url, download=False)
        if "entries" in info:
            self._video_urls = [entry["webpage_url"] for entry in info["entries"]]
        else:
            self._video_urls = [info["webpage_url"]]
        self._url_idx = 0

        # Add a hook to dynamically determine what the output filename is for
        # each video.
        ydl_opts = {
            "format": "bestvideo+bestaudio/best",
            "merge_output_format": "mp4",
            "outtmpl": str(output_directory / "%(id)s.%(ext)s")
        }
        self._youtube_dl = yt_dlp.YoutubeDL(ydl_opts)

    def read(self) -> Dict[str, str]:

        if self._url_idx == len(self._video_urls):
            return None
        url = self._video_urls[self._url_idx]

        meta = self._youtube_dl.extract_info(url, download=False)
        meta = self._youtube_dl.sanitize_info(meta)
        path = self._youtube_dl.prepare_filename(meta)

        self._youtube_dl.download(url)
        self._url_idx += 1

        return {"data": path}







__all__ = [
    "AudioVectorizer"
]

from abc import abstractmethod
from typing import Any, List, Tuple, Union

import numpy as np

from radient.tasks.vectorizers._base import Vectorizer
from radient.utils import fully_qualified_name
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector

librosa = LazyImport("librosa")


class AudioVectorizer(Vectorizer):

    @abstractmethod
    def __init__(self):
        super().__init__()

    def _preprocess(self, audio: Any, **kwargs) -> np.ndarray:
        if isinstance(audio, tuple) and isinstance(audio[0], np.ndarray):
            waveform, source_rate = audio
        elif isinstance(audio, str):
            waveform, source_rate = librosa.load(audio, sr=None, mono=False)
            if len(waveform.shape) == 1:
                waveform = np.expand_dims(waveform, 0)

        if source_rate != self.sample_rate:
            waveform = librosa.resample(
                waveform,
                orig_sr=source_rate,
                target_sr=self.sample_rate
            )

        return waveform

    @property
    @abstractmethod
    def sample_rate(self) -> int:
        """Returns the sample rate required by this vectorizer.
        """
        pass
    


__all__ = [
    "CountTextVectorizer"
]

from typing import Dict, List, Optional
import warnings

from radient.tasks.vectorizers.text._base import TextVectorizer
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector

CountVectorizer = LazyImport("sklearn.feature_extraction.text", attribute="CountVectorizer", package_name="scikit-learn")
TfidfVectorizer = LazyImport("sklearn.feature_extraction.text", attribute="TfidfVectorizer", package_name="scikit-learn")
HashingVectorizer = LazyImport("sklearn.feature_extraction.text", attribute="HashingVectorizer", package_name="scikit-learn")


class SklearnTextVectorizer(TextVectorizer):
    """Text vectorization with `sentence-transformers`.
    """

    def __init__(self, **kwargs):
        super().__init__()
        self._model = CountVectorizer(**kwargs)

    def _vectorize(self, texts: str, **kwargs) -> Vector:
        vectors = self._model.transform(texts)
        # TODO(fzliu): sparse vector type
        return vectors
        #return [v.view(Vector) for v in vectors]



__all__ = [
    "TextVectorizer",
    "CohereTextVectorizer"
    "ImageBindTextVectorizer"
    "SBERTTextVectorizer",
    "SklearnTextVectorizer",
    "VoyageTextVectorizer"
    "text_vectorizer"
]

from typing import Optional

from radient.tasks.vectorizers.text._base import TextVectorizer
from radient.tasks.vectorizers.text.cohere import CohereTextVectorizer
from radient.tasks.vectorizers.text.imagebind import ImageBindTextVectorizer
from radient.tasks.vectorizers.text.sbert import SBERTTextVectorizer
from radient.tasks.vectorizers.text.sklearn import SklearnTextVectorizer
from radient.tasks.vectorizers.text.voyage import VoyageTextVectorizer


def text_vectorizer(method: str = "sbert", **kwargs) -> TextVectorizer:
    """Creates a text vectorizer specified by `method`.
    """

    # Return a reasonable default vectorizer in the event that the user does
    # not specify one.
    if method in ("sbert", "sentence-transformers"):
        return SBERTTextVectorizer(**kwargs)
    elif method in ("imagebind",):
        return ImageBindTextVectorizer(**kwargs)
    elif method in ("sklearn", "scikit-learn"):
        return SklearnTextVectorizer(**kwargs)
    elif method in ("cohere",):
        return CohereTextVectorizer(**kwargs)
    elif method in ("voyage", "voyageai"):
        return VoyageTextVectorizer(**kwargs)
    else:
        raise NotImplementedError



# Vector base class
from radient.vector import Vector

# Vectorization only
from radient.tasks.vectorizers import (
    audio_vectorizer,
    graph_vectorizer,
    image_vectorizer,
    molecule_vectorizer,
    text_vectorizer,
    multimodal_vectorizer
)

# Orchestration
from radient.factory import make_operator
from radient.orchestrate.runners import (
    LocalRunner,
    LazyLocalRunner
)
from radient.orchestrate.workflow import Workflow


import urllib.request
from pathlib import Path

from typing import Any, Optional


def download_cache_file(url: str, filename: Optional[str] = None) -> None:
    """Download a file from a URL and save it to a local file.
    """
    if not filename:
        filename = url.split("/")[-1].split("?")[0]
    path = Path.home() / ".cache" / "radient" / filename
    path.parents[0].mkdir(parents=True, exist_ok=True)
    if not path.exists():
        urllib.request.urlretrieve(url, path)
    return path


def fully_qualified_name(instance: Any) -> str:
    return f"{instance.__class__.__module__}.{instance.__class__.__qualname__}"


from radient.tasks.vectorizers._base import Vectorizer
from radient.tasks.vectorizers.audio import audio_vectorizer
from radient.tasks.vectorizers.graph import graph_vectorizer
from radient.tasks.vectorizers.image import image_vectorizer
from radient.tasks.vectorizers.molecule import molecule_vectorizer
from radient.tasks.vectorizers.text import text_vectorizer
from radient.tasks.vectorizers.multimodal import multimodal_vectorizer



from pathlib import Path

from typing import Dict, List, Optional, Tuple, Union

from radient._milvus import _MilvusInterface
from radient.tasks.sinks._base import Sink
from radient.utils import fully_qualified_name
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector


DEFAULT_MILVUS_URI = str(Path.home() / ".radient" / "default.db")
#DEFAULT_MILVUS_URI = "http://127.0.0.1:19530"
DEFAULT_COLLECTION_NAME = "radient"


class MilvusSink(Sink):

    def __init__(
        self,
        operation: str,
        milvus_uri: str = DEFAULT_MILVUS_URI,
        collection_name: str = DEFAULT_COLLECTION_NAME,
        vector_field: Optional[str] = None,
        **kwargs
    ):
        super().__init__()
        self._operation = operation
        self._milvus_uri = milvus_uri
        self._collection_name = collection_name
        self._vector_field = vector_field

    def transact(
        self,
        vectors: Union[Vector, List[Vector]],
        **kwargs
    ) -> Dict[str, Union[int, List[int]]]:
        if not isinstance(vectors, list):
            vectors = [vectors]
        client, info = _MilvusInterface._get_client(
            milvus_uri=self._milvus_uri,
            collection_name=self._collection_name,
            dimension=vectors[0].size
        )
        # If `field_name` is None, attempt to automatically acquire the field
        # name from the collection info.
        vector_field = self._vector_field or info["dense"]

        if self._operation == "insert":
            return client.insert(
                collection_name=self._collection_name,
                data=[v.todict(vector_field=vector_field) for v in vectors],
                **kwargs
            )
        elif self._operation == "search":
            return client.search(
                collection_name=self._collection_name,
                data=[v.tolist() for v in vectors],
                **kwargs
            )
        else:
            raise TypeError("invalid Milvus operation")



__all__ = [
    "MoleculeVectorizer"
]

from abc import abstractmethod
from typing import Any, List

from radient.vector import Vector
from radient.utils import fully_qualified_name
from radient.utils.lazy_import import LazyImport
from radient.tasks.vectorizers._base import Vectorizer

Chem = LazyImport("rdkit.Chem")


class MoleculeVectorizer(Vectorizer):

    @abstractmethod
    def __init__(self):
        super().__init__()

    def _preprocess(self, molecule: Any, **kwargs) -> str:
        full_name = fully_qualified_name(molecule)
        if full_name == "builtins.str":
            return molecule
        elif full_name == "rdkit.Chem.rdchem.Mol":
            # TODO: don't do this step for `RDKitMoleculeVectorizer`
            return Chem.MolToSmiles(molecule)
        else:
            raise TypeError



__all__ = [
    "TimmImageVectorizer",
    "image_vectorizer"
]

from typing import Optional

from radient.tasks.vectorizers.image._base import ImageVectorizer
from radient.tasks.vectorizers.image.timm import TimmImageVectorizer
from radient.tasks.vectorizers.image.imagebind import ImageBindImageVectorizer


def image_vectorizer(method: str = "timm", **kwargs) -> ImageVectorizer:
    """Creates an image vectorizer specified by `method`.
    """

    # Return a reasonable default vectorizer in the event that the user does
    # not specify one.
    if method in ("timm", "pytorch-image-models"):
        return TimmImageVectorizer(**kwargs)
    elif method in ("imagebind",):
        return ImageBindImageVectorizer(**kwargs)
    else:
        raise NotImplementedError



from radient.utils.lazy_import import LazyImport

imagebind_model = LazyImport("imagebind.models", attribute="imagebind_model", package_name="git+https://github.com/fzliu/ImageBind@main")

IMAGEBIND_MODULE_NAMES = (
    "modality_preprocessors",
    "modality_trunks",
    "modality_heads",
    "modality_postprocessors"
)


def create_imagebind_model(modality: str, model_name: str = "imagebind_huge"):
    """Wrapper around `imagebind_model` to load a specific modality. Modalities
    aside from the one specified are removed from the model. (It's might be
    better to get this code merged into the original ImageBind repo.)
    """

    if modality == "image":
        modality = imagebind_model.ModalityType.VISION
    elif modality == "text":
        modality = imagebind_model.ModalityType.TEXT
    elif modality == "audio":
        modality = imagebind_model.ModalityType.AUDIO

    model = getattr(imagebind_model, model_name)(pretrained=True)

    # Delete unnecessary modality trunks, preprocessors, and postprocessors
    # from the model.
    for module_name in IMAGEBIND_MODULE_NAMES:
        for modality_type in imagebind_model.ModalityType.__dict__.values():
            if modality_type != modality:
                module = getattr(model, module_name)
                delattr(module, modality_type)

    return model




import importlib
import importlib.util
import subprocess
from types import ModuleType
from typing import Any, List, Optional
import warnings

import pip


def prompt_install(
    package: str,
    version: Optional[str] = None
) -> bool:
    """Checks whether the user wants to install a module before proceeding.
    """
    if version:
        package = f"{package}>={version}"
    # Ignore "no" responses in `prompt_install` so the user can optionally
    # install it themselves when prompted.
    if input(f"This requires {package}. Install? [Y/n]\n") == "Y":
        if subprocess.check_call(["pip", "install", "-q", package]) == 0:
            return True
        else:
            warnings.warn(f"Could not install required package {package}")
    return False


class LazyImport(ModuleType):
    """Lazily import a module to avoid unnecessary dependencies. If a required
    dependency does not exist, it will prompt the user for it.

    Adapted from tensorflow/python/util/lazy_loader.py.
    """

    def __init__(
        self,
        name: str,
        attribute: Optional[str] = None,
        package_name: Optional[str] = None,
        min_version: Optional[str] = None
    ):
        super().__init__(name)
        self._attribute = attribute
        self._top_name = name.split(".")[0]
        self._package_name = package_name if package_name else self._top_name
        self._min_version = min_version
        self._module = None

    def __call__(self, *args, **kwargs) -> Any:
        return self._evaluate()(*args, **kwargs)

    def __getattr__(self, attribute: str) -> Any:
        return getattr(self._evaluate(), attribute)

    def __dir__(self) -> List:
        return dir(self._evaluate())

    def _evaluate(self) -> ModuleType:
        if not self._module:
            if not importlib.util.find_spec(self._top_name):
                prompt_install(self._package_name, self._min_version)
        self._module = importlib.import_module(self.__name__)
        if self._min_version and self._module.__version__ < self._min_version:
            prompt_install(self._package_name, self._min_version)
            self._module = importlib.import_module(self.__name__)
        if self._attribute:
            return getattr(self._module, self._attribute)
        return self._module



__all__ = [
    "ImageBindTextVectorizer"
]

from typing import Any, List

import numpy as np

import urllib.request

from radient.tasks.accelerate import export_to_onnx, ONNXForward
from radient.tasks.vectorizers._imagebind import create_imagebind_model
from radient.tasks.vectorizers._imagebind import imagebind_model
from radient.tasks.vectorizers.text._base import TextVectorizer
from radient.utils import download_cache_file
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector

SimpleTokenizer = LazyImport("imagebind.models.multimodal_preprocessors", attribute="SimpleTokenizer", package_name="git+https://github.com/fzliu/ImageBind@main")
torch = LazyImport("torch")

IMAGEBIND_VOCAB_URL = "https://github.com/fzliu/ImageBind/raw/main/imagebind/bpe/bpe_simple_vocab_16e6.txt.gz"



class ImageBindTextVectorizer(TextVectorizer):
    """Computes image embeddings using FAIR's ImageBind model.
    """

    def __init__(self, model_name = "imagebind_huge", **kwargs):
        super().__init__()
        self._model_name = model_name
        self._model = create_imagebind_model(model_name=model_name, modality="text")
        self._model.eval()
        vocab_path = download_cache_file(IMAGEBIND_VOCAB_URL)
        self._tokenizer = SimpleTokenizer(bpe_path=vocab_path)

    def _vectorize(self, text: str, **kwargs) -> Vector:
        # TODO(fzliu): dynamic batching
        with torch.inference_mode():
            tokens = self._tokenizer(text).unsqueeze(0)
            output = self._model({imagebind_model.ModalityType.TEXT: tokens})
            vector = output[imagebind_model.ModalityType.TEXT].squeeze()
            if isinstance(vector, torch.Tensor):
                vector = vector.numpy()
        return vector.view(Vector)

    def accelerate(self):
        modality = imagebind_model.ModalityType.TEXT
        input_names = output_names = [modality]
        inputs = ({modality: self._tokenizer("a")}, {})
        onnx_model_path = export_to_onnx(
            self,
            inputs,
            axes_names=["batch_size"],
            input_names=input_names,
            output_names=output_names,
            model_type="pytorch"
        )

        self._model.forward = ONNXForward(
            onnx_model_path,
            output_names=output_names,
        )

    @property
    def sample_rate(self):
        return 16_000
    




from abc import abstractmethod
from typing import Any

from radient.tasks._base import Task


class Source(Task):
    """Sources in Radient are task objects that yield data. Depending on the
    downstream transform, this can be raw bytes, or it can be filenames/URIs.
    """

    @abstractmethod
    def __init__(self):
        super().__init__()

    def __call__(self, *args, **kwargs):
        return self.read(*args, **kwargs)

    @abstractmethod
    def read(self, **kwargs) -> Any:
        pass



from abc import abstractmethod
from typing import Any

import numpy as np

from radient.tasks._base import Task


class Transform(Task):
    """Transforms are operations that perform multimodal data transformation,
    such as such as turning a video into independent frames. Because these are
    usually I/O-bound operations, batching is not innately supported.
    """

    @abstractmethod
    def __init__(self):
        super().__init__()

    def __call__(self, *args, **kwargs):
        return self.transform(*args, **kwargs)

    @abstractmethod
    def transform(self, data: Any) -> Any:
        pass



__all__ = [
    "ImageBindImageVectorizer"
]

from typing import Any, List

from radient.tasks.accelerate import export_to_onnx, ONNXForward
from radient.tasks.vectorizers._imagebind import create_imagebind_model
from radient.tasks.vectorizers._imagebind import imagebind_model
from radient.tasks.vectorizers.image._base import ImageVectorizer
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector

Image = LazyImport("PIL", attribute="Image", package_name="Pillow")
torch = LazyImport("torch")
transforms = LazyImport("torchvision", attribute="transforms")


class ImageBindImageVectorizer(ImageVectorizer):
    """Computes image embeddings using FAIR's ImageBind model.
    """

    def __init__(self, model_name: str = "imagebind_huge", **kwargs):
        super().__init__()
        self._model_name = model_name
        # TODO(fzliu): remove non-image trunks from this model
        self._model = create_imagebind_model(model_name=model_name, modality="image")
        self._model.eval()
        self._transform = transforms.Compose([
            transforms.Resize(
                224, interpolation=transforms.InterpolationMode.BICUBIC
            ),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize(
                mean=(0.48145466, 0.4578275, 0.40821073),
                std=(0.26862954, 0.26130258, 0.27577711),
            ),
        ])

    def _vectorize(self, image: Image, **kwargs) -> Vector:
        # TODO(fzliu): dynamic batching
        with torch.inference_mode():
            x = self._transform(image.convert("RGB")).unsqueeze(0)
            output = self._model({imagebind_model.ModalityType.VISION: x})
            vector = output[imagebind_model.ModalityType.VISION].squeeze()
            if isinstance(vector, torch.Tensor):
                vector = vector.numpy()
        return vector.view(Vector)

    def accelerate(self):
        modality = imagebind_model.ModalityType.VISION
        inputs = ({modality: torch.randn((1, 3, 224, 224))}, {})
        input_names = output_names = [modality]
        onnx_model_path = export_to_onnx(
            self,
            inputs,
            axes_names=["batch_size"],
            input_names=input_names,
            output_names=output_names,
            model_type="pytorch"
        )
        # TODO(fzliu): delete all tensors from model
        self._model.forward = ONNXForward(
            onnx_model_path,
            output_names=output_names,
        )




from collections import OrderedDict
from typing import Any, Dict, List, Sequence, Optional, Union

import numpy as np

from radient._milvus import _MilvusInterface


class Vector(np.ndarray):
    """Wrapper around `numpy.ndarray` specifically for working with embeddings.
    We try to use Numpy naming conventions here where possible, such as concise
    function names and 
    """

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    def __array_finalize__(self, obj):
        """Attach metadata to be associated with this vector.
        """
        self._metadata = OrderedDict()

    @property
    def metadata(self) -> OrderedDict:
        return self._metadata

    @metadata.setter
    def metadata(self, data: Dict):
        self._metadata = OrderedDict(data)

    def putmeta(
        self,
        key: str,
        value: Union[Dict[str, Union[str, float, int]], str, float, int]
    ) -> "Vector":
        self._metadata[key] = value
        # Enable chaining function calls together.
        return self

    def popmeta(
        self,
        key: str
    ) -> Union[Dict[str, Union[str, float, int]], str, float, int]:
        return self._metadata.pop(key)

    def todict(
        self,
        vector_field: str = "vector"
    ) -> Dict[str, Union["Vector", str, float, int]]:
        return dict(self._metadata, **{vector_field: self.tolist()})

    def store(
        self,
        sink_type: Union[Sequence[str], str] = "vectordb",
        **kwargs
    ):
        """Stores this vector in the specified sink. This function is for
        convenience only.
        """
        if isinstance(sink_type, str):
            sink_type = [sink_type]
        for sink in sink_type:
            if sink == "vectordb":
                return self._store_vectordb(**kwargs)

    def _store_vectordb(
        self,
        milvus_uri: str,
        collection_name: str,
        field_name: Optional[str] = None
    ) -> Dict[str, Union[str, List]]:
        """Stores this vector in the specified collection.
        """
        client, info = _MilvusInterface(milvus_uri, collection_name, dim=self.size)
        field = info["dense"]
        # We can get away with using the dict constructor because all schema
        # field names are strings.
        data = dict(self._metadata, **{field: self.tolist()})
        return client.insert(
            collection_name=collection_name,
            data=data
        )

    def _search_vectordb(
        self,
        milvus_uri: str,
        collection_name: str,
        metric_type: Optional[str] = "COSINE",
        topk: int = 10
    ) -> List[List[np.ndarray]]:
        """Queries the specified collection for nearest neighbors.
        """
        client, info = _MilvusInterface(milvus_uri, collection_name)
        return client.search(
            collection_name="test_collection",
            data=[self.tolist()],
            limit=topk,
            search_params={"metric_type": metric_type, "params": {}}
        )




from collections.abc import Iterator
from pathlib import Path
from typing import Dict

from radient.tasks.sources._base import Source


def _path_walk(path):
    p = Path(path)
    if p.is_file():
        yield path
    elif p.is_dir():
        for sub in p.iterdir():
            yield from _path_walk(sub)


class LocalSource(Source):
    """Reads filenames from a local directory. This source is mostly useful for
    backfills from local disk and or for testing purposes.
    """

    def __init__(self, path: str, **kwargs):
        super().__init__()
        self._paths_iter = _path_walk(path)

    def read(self) -> Dict[str, str]:
        try:
            return {"data": self._paths_iter.__next__()}
        except StopIteration:
            return None



__all__ = [
    "GraphVectorizer"
]

from abc import abstractmethod
from typing import Any

import numpy as np

from radient.tasks.vectorizers._base import Vectorizer
from radient.utils import fully_qualified_name
from radient.utils.lazy_import import LazyImport

nx = LazyImport("networkx")
sp = LazyImport("scipy")


class GraphVectorizer(Vectorizer):

    @abstractmethod
    def __init__(self):
        super().__init__()

    @classmethod
    def _preprocess(cls, graph: Any, **kwargs) -> str:
        # Turn input graphs into adjacency matrices.
        full_name = fully_qualified_name(graph)
        if isinstance(graph, sp.sparse.sparray):
            return graph
        elif isinstance(graph, sp.sparse.spmatrix):
            # TODO(fzliu): determine if support for this is necessary
            raise NotImplementedError
        elif full_name == "networkx.classes.graph.Graph":
            return nx.to_scipy_sparse_array(graph)
        else:
            raise TypeError



from collections import defaultdict, OrderedDict
from collections.abc import Callable, Iterator
from typing import Any, Dict, List, Optional, Sequence, Union
from graphlib import TopologicalSorter

from radient.utils.flatten_inputs import flattened


class Workflow:
    """Workflows are used to chain together independent tasks together in a
    DAG. The output of each task is maintained in a table and passed to
    subsequent tasks that need the corresponding result.
    """

    def __init__(self):
        self._runners = OrderedDict()
        self._dependencies = {}
        self._runner_graph = None

    def __call__(self, *args, **kwargs) -> Any:
        self.compile()
        return self.execute(*args, **kwargs)

    def add(
        self,
        runner: Callable,
        name: str,
        dependencies: Optional[Sequence[str]] = None
    ) -> "Workflow":

        # By default, new tasks have a single dependency: the preceding task.
        if not dependencies:
            names = list(self._runners.keys())
            dependencies = (names[-1],) if names else ()
        self._dependencies[name] = dependencies

        self._runners[name] = runner

        return self

    def compile(self):
        self._runner_graph = TopologicalSorter(self._dependencies)
        self._all_outputs = defaultdict(list)

    def execute(
        self,
        extra_vars: Optional[Dict[str, Dict[str, Any]]] = None,
        **kwargs
    ) -> Any:
        if self._runner_graph is None:
            raise ValueError("call compile() first")

        extra_vars = extra_vars or {}

        # TODO(fzliu): workflows may be persistent rather than returning a
        # single output or set of outputs
        for name in self._runner_graph.static_order():
            inputs = []
            if not self._dependencies[name]:
                # A task with no dependencies is a "seed" task.
                inputs.append([kwargs])
            else:
                for d in self._dependencies[name]:
                    inputs.append(self._all_outputs[d][-1])

            # A task can return a single item or multiple items in a list.
            outputs = []
            for args, _ in flattened(*inputs):
                kwargs = {k: v for d in args for k, v in d.items()}
                kwargs.update(extra_vars.get(name, {}))
                result = self._runners[name](**kwargs)
                if isinstance(result, list):
                    outputs.extend(result)
                else:
                    outputs.append(result)
            self._all_outputs[name].append(outputs)

        return self._all_outputs[name]





__all__ = [
    "TorchaudioAudioVectorizer",
    "audio_vectorizer"
]

from typing import Optional

from radient.tasks.vectorizers.audio._base import AudioVectorizer
from radient.tasks.vectorizers.audio.imagebind import ImageBindAudioVectorizer
from radient.tasks.vectorizers.audio.torchaudio import TorchaudioAudioVectorizer


def audio_vectorizer(method: str = "torchaudio", **kwargs) -> AudioVectorizer:
    """Creates an image vectorizer specified by `method`.
    """

    # Return a reasonable default vectorizer in the event that the user does
    # not specify one.
    if method in ("torchaudio",):
        return TorchaudioAudioVectorizer(**kwargs)
    elif method in ("imagebind",):
        return ImageBindAudioVectorizer(**kwargs)
    else:
        raise NotImplementedError



__all__ = [
    "VideoDemuxTransform"
]

from typing import Optional

from radient.tasks.transforms.video_demux._base import VideoDemuxTransform
from radient.tasks.transforms.video_demux.default import DefaultVideoDemuxTransform
from radient.tasks.transforms.video_demux.ffmpeg import FFmpegVideoDemuxTransform


def video_demux_transform(method: str = "default", **kwargs) -> VideoDemuxTransform:
    """Creates a video demultiplexer specified by `method`.
    """

    # Return a reasonable default vectorizer in the event that the user does
    # not specify one.
    if method in ("default", None):
        return DefaultVideoDemuxTransform(**kwargs)
    elif method in ("ffmpeg",):
        return FFmpegVideoDemuxTransform(**kwargs)
    else:
        raise NotImplementedError



__all__ = [
    "SBERTTextVectorizer"
]

from typing import List

from radient.tasks.accelerate import export_to_onnx, ONNXForward
from radient.tasks.vectorizers.text._base import TextVectorizer
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector

SentenceTransformer = LazyImport("sentence_transformers", attribute="SentenceTransformer", package_name="sentence-transformers")
torch = LazyImport("torch")


class SBERTTextVectorizer(TextVectorizer):
    """Text vectorization with `sentence-transformers`.
    """

    def __init__(self, model_name: str = "BAAI/bge-small-en-v1.5", **kwargs):
        super().__init__()
        if "model_name_or_path" in kwargs:
            model_name = kwargs.pop("model_name_or_path")
        self._model_name = model_name
        self._model = SentenceTransformer(model_name_or_path=model_name, **kwargs)

    def _vectorize(self, text: str, **kwargs) -> Vector:
        # TODO(fzliu): token length check
        # TODO(fzliu): dynamic batching
        with torch.inference_mode():
            vector = self._model.encode(text)
        return vector.view(Vector)

    @property
    def model_name(self):
        return self._model_name

    def accelerate(self):
        # Store the model in ONNX format to maximize compatibility with
        # different backends. Since `sentence-transformers` takes a single
        # dictionary input in its underlying `forward` call, the export
        # function will need to take a second empty dictionary as kwargs.
        # Output names are acquired by running the `encode` function and
        # specifying all outputs.
        model_args = (self._model.tokenize(["a"]), {})
        input_names = list(model_args[0].keys())
        output_names = list(self._model.encode("a", output_value=None).keys())
        onnx_model_path = export_to_onnx(
            self,
            model_args,
            axes_names=["batch_size", "seq_len"],
            input_names=input_names,
            output_names=output_names,
            model_type="pytorch"
        )

        # Monkey-patch the the underlying model's `forward` function to run the
        # optimized ONNX model rather than the torch version.
        self._model.forward = ONNXForward(
            onnx_model_path,
            output_names=output_names,
            output_class=torch.tensor,
        )



__all__ = [
    "ImageBindAudioVectorizer"
]

from typing import Any, List

import numpy as np

from radient.tasks.accelerate import export_to_onnx, ONNXForward
from radient.tasks.vectorizers._imagebind import create_imagebind_model
from radient.tasks.vectorizers._imagebind import imagebind_model
from radient.tasks.vectorizers.audio._base import AudioVectorizer
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector

torch = LazyImport("torch")
transforms = LazyImport("torchvision", attribute="transforms")
waveform2melspec = LazyImport("imagebind.data", attribute="waveform2melspec")

CLIP_DURATION = 2
NUM_MEL_BINS = 128
TARGET_LENGTH = 204


class ImageBindAudioVectorizer(AudioVectorizer):
    """Computes image embeddings using FAIR's ImageBind model.
    """

    def __init__(self, model_name = "imagebind_huge", **kwargs):
        super().__init__()
        self._model_name = model_name
        self._model = create_imagebind_model(model_name=model_name, modality="audio")
        self._model.eval()
        self._normalize = transforms.Normalize(mean=-4.268, std=9.138)

    def _transform(self, waveform: np.ndarray, **kwargs):
        output = []
        # Split the waveform into clips of duration CLIP_DURATION. Each
        # waveform is then converted into its Mel spectrum representation.
        waveform = torch.from_numpy(waveform)
        samples_per_clip = self.sample_rate * CLIP_DURATION
        for n in np.arange(0, waveform.shape[1], samples_per_clip):
            end = n + samples_per_clip
            mel_spec = waveform2melspec(
                waveform[:,n:end],
                self.sample_rate,
                NUM_MEL_BINS,
                TARGET_LENGTH
            )
            output.append(self._normalize(mel_spec))
        return torch.stack(output, dim=0)

    def _preprocess(self, audio: Any, **kwargs) -> np.ndarray:
        audio = super()._preprocess(audio)
        audio = self._transform(audio).unsqueeze(0)
        return audio

    def _vectorize(self, audio: np.ndarray, **kwargs) -> Vector:
        # TODO(fzliu): dynamic batching
        with torch.inference_mode():
            output = self._model({imagebind_model.ModalityType.AUDIO: audio})
            vector = output[imagebind_model.ModalityType.AUDIO].squeeze()
            if isinstance(vector, torch.Tensor):
                vector = vector.numpy()
        return vector.view(Vector)

    def accelerate(self):
        modality = imagebind_model.ModalityType.AUDIO
        inputs = ({modality: torch.randn((1, 400))}, {})
        input_names = output_names = [modality]
        onnx_model_path = export_to_onnx(
            self,
            inputs,
            axes_names=["batch_size", "seq_len"],
            input_names=[modality],
            output_names=[modality],
            model_type="pytorch"
        )

        self._model.forward = ONNXForward(
            onnx_model_path,
            output_names=output_names,
        )

    @property
    def sample_rate(self):
        return 16_000
    




from abc import ABC, abstractmethod
import random
from typing import Dict, Tuple, Type, Optional

from radient.utils.flatten_inputs import flattened


class Runner(ABC):

    @abstractmethod
    def __init__(
        self,
        task: Type,
        task_params: Optional[Dict] = None,
        flatten_inputs: Optional[str] = False
    ):
        self._task = task
        self._task_params = task_params or {}
        self._flatten_inputs = flatten_inputs
        self._result = None

    def _evaluate(self):
        self._result = self._task(
            **self._task_params
        )

    @property
    def result(self):
        return self._result

    def __call__(self, *args, **kwargs):
        if self._flatten_inputs:
            outputs = []
            for flat_args, flat_kwargs in flattened(*args, **kwargs):
                outputs.append(self.result(*flat_args, **flat_kwargs))
            return outputs
        return self.result(*args, **kwargs)


class LocalRunner(Runner):
    """Evaluate a function or instance locally.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._evaluate()


class LazyLocalRunner(Runner):
    """Lazily (on-demand) evaluate a function or instance.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def result(self):
        if not self._result:
            self._evaluate()
        return self._result



__all__ = [
    "RDKitMoleculeVectorizer"
]

from typing import Any, List

import numpy as np

from radient.utils.lazy_import import LazyImport
from radient.vector import Vector
from radient.tasks.vectorizers.molecule._base import MoleculeVectorizer

Chem = LazyImport("rdkit.Chem")
AllChem = LazyImport("rdkit.Chem.AllChem")


class RDKitMoleculeVectorizer(MoleculeVectorizer):
    """Generates binary fingerprints for molecules.
    """

    def __init__(self, fingerprint_type: str = "topological", **kwargs):
        super().__init__()
        self._fingerprint_type = fingerprint_type
        if fingerprint_type == "topological":
            self._fpgen = AllChem.GetRDKitFPGenerator(**kwargs)
        elif fingerprint_type == "morgan":
            self._fpgen = AllChem.GetMorganGenerator(**kwargs)

    def _vectorize(self, molecule: str, **kwargs) -> Vector:
        if isinstance(molecule, str):
            molecule = Chem.MolFromSmiles(molecule)
        fp = self._fpgen.GetFingerprint(molecule)
        # Use dtype=bool to avoid having to bit-pack into `uint8`.
        vector = np.array(fp.ToList(), dtype=bool)
        return vector.view(Vector)

    @property
    def fingerprint_type(self) -> str:
        return self._fingerprint_type



__all__ = [
    "FastRPGraphVectorizer",
    "graph_vectorizer"
]

from typing import Optional

from radient.tasks.vectorizers.graph._base import GraphVectorizer
from radient.tasks.vectorizers.graph.fastrp import FastRPGraphVectorizer


def graph_vectorizer(method: str = "fastrp", **kwargs) -> GraphVectorizer:
    """Creates an image vectorizer specified by `method`.
    """

    # Return a reasonable default vectorizer in the event that the user does
    # not specify one.
    if method in ("fastrp"):
        return FastRPGraphVectorizer(**kwargs)
    else:
        raise NotImplementedError



from abc import abstractmethod
from pathlib import Path
from typing import Dict, List
import uuid

from radient.tasks.transforms._base import Transform


class VideoDemuxTransform(Transform):

    def __init__(self,
        interval: float = 2.0,
        output_directory: str = "~/.radient/data/video_demux",
        **kwargs
    ):
        super().__init__()
        self._interval = interval
        output_directory = Path(output_directory).expanduser()
        self._output_directory = output_directory

    @abstractmethod
    def transform(self, data: str) -> Dict[str, List[str]]:
        """Extracts frames and audio snippets from a video file.
        """
        pass

    def _make_output_dir(self):
        # The full output directory comes from a combination of the user's
        # specification plus a unique identifier for the current run.
        # TODO(fzliu): switch to an incremental identifier e.g. UUIDv7
        output_path = Path(self._output_directory) / str(uuid.uuid4())
        output_path.mkdir(parents=True, exist_ok=True)
        return output_path



__all__ = [
    "ImageVectorizer"
]

from abc import abstractmethod
import base64
import io
from pathlib import Path
from typing import Any, List
import urllib.request

import numpy as np

from radient.tasks.vectorizers._base import Vectorizer
from radient.utils import fully_qualified_name
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector

Image = LazyImport("PIL.Image", package_name="pillow")
validators = LazyImport("validators")


class ImageVectorizer(Vectorizer):

    @abstractmethod
    def __init__(self):
        super().__init__()

    def _preprocess(self, image: Any, mode: str = "RGB", **kwargs) -> "Image.Image":
        """Converts the input images into a common format, i.e. a PIL Image.
        Video files are loaded into `torchvision.io.VideoReader` objects.
        """
        # Acquire the full class path, i.e. qualified name plus module name.
        # There might be a better way to do this that takes module rebinding
        # into consideration, but this will do for now.
        full_name = fully_qualified_name(image)
        if full_name == "PIL.Image.Image":
            return image
        elif full_name == "numpy.ndarray":
            return Image.fromarray(image, mode=mode)
        elif full_name == "builtins.str":
            # For string inputs, we support three options - a base64 encoded
            # string containing the image data, a path to a filename which is
            # a valid image type, or a URL that contains the image.
            imgpath = Path(image)
            if imgpath.suffix in Image.registered_extensions().keys():
                if imgpath.exists():
                    return Image.open(image)
                elif validators.url(image):
                    with urllib.request.urlopen(image) as resp:
                        imgbytes = io.BytesIO(resp.read())
                    return Image.open(imgbytes)
            else:
                try:
                    imgbytes = io.BytesIO(base64.b64decode(image))
                    return Image.open(imgbytes)
                except:
                    raise TypeError
        else:
            raise TypeError



from typing import Dict, List
import shutil
import subprocess

import numpy as np

from radient.tasks.transforms.video_demux._base import VideoDemuxTransform


class FFmpegVideoDemuxTransform(VideoDemuxTransform):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def transform(self, data: str) -> Dict[str, List[str]]:
        """Extracts frames and audio snippets from a video file.
        """

        # Ensure that FFmpeg is installed and available from the command line.
        if not shutil.which("ffmpeg"):
            raise FileNotFoundError(f"ffmpeg not found, try specifying 'method': 'default' in params")
        if not shutil.which("ffprobe"):
            raise FileNotFoundError(f"ffmpeg not found, try specifying 'method': 'default' in params")

        frames = {"data": [], "modality": "image"}
        audios = {"data": [], "modality": "audio"}
        video_path = data
        output_path = self._make_output_dir()

        # Grab video information using ffprobe.
        frame_info = subprocess.run(
            ["ffprobe", "-v", "error", "-select_streams", "v:0",
             "-count_packets", "-show_entries",
             "stream=r_frame_rate,nb_read_packets", "-of",
             "default=noprint_wrappers=1:nokey=1", video_path],
            capture_output=True, text=True).stdout.split()
        frame_rate = eval(frame_info[0])
        frame_count = eval(frame_info[1])
        frame_interval = frame_rate * self._interval
        audio_info = subprocess.run(
            ["ffprobe", "-v", "error", "-select_streams", "a:0",
             "-show_entries", "stream=sample_rate", "-of",
             "default=noprint_wrappers=1:nokey=1", video_path],
            capture_output=True, text=True).stdout.split()

        for i, n in enumerate(np.arange(0, frame_count, frame_interval)):
            start_time = n / frame_rate

            # Extract frames.
            frame_path = str(output_path / f"frame_{i:04d}.png")
            subprocess.run(["ffmpeg", "-v", "error", "-ss", str(start_time),
                            "-i", video_path, "-vframes", "1", frame_path])
            frames["data"].append(frame_path)

            # Extract audio.
            audio_path = str(output_path / f"audio_{i:04d}.wav")
            subprocess.run(["ffmpeg", "-v", "error", "-ss", str(start_time),
                            "-i", video_path, "-t", str(self._interval),
                            "-q:a", "0", "-map", "a", audio_path])
            audios["data"].append(audio_path)

        return [frames, audios]



from typing import Dict, List

import numpy as np

from radient.tasks.transforms.video_demux._base import VideoDemuxTransform
from radient.utils.lazy_import import LazyImport

cv2 = LazyImport("cv2", package_name="opencv-python")
librosa = LazyImport("librosa")
sf = LazyImport("soundfile")


class DefaultVideoDemuxTransform(VideoDemuxTransform):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def transform(self, data: str) -> Dict[str, List[str]]:
        """Extracts frames and audio snippets from a video file.
        """

        video_path = data
        output_path = self._make_output_dir()

        # Grab the total number of frames as well as the video's FPS to
        # determine the interval in frames and stopping condition.
        video_capture = cv2.VideoCapture(video_path)
        frame_count = video_capture.get(cv2.CAP_PROP_FRAME_COUNT)
        frame_interval = video_capture.get(cv2.CAP_PROP_FPS) * self._interval

        frames = {"data": [], "modality": "image"}
        for i, n in enumerate(np.arange(0, frame_count, frame_interval)):
            video_capture.set(cv2.CAP_PROP_POS_FRAMES, int(n))
            retval, frame = video_capture.read()
            if not retval:
                break
            frame_path = str(output_path / f"frame_{i:04d}.png")
            cv2.imwrite(frame_path, frame)
            frames["data"].append(frame_path)

        video_capture.release()

        # Extract audio snippet as raw data. With the raw audio, we store it
        # in `.wav` format with the original sample rate.
        audios = {"data": [], "modality": "audio"}
        waveform, sample_rate = librosa.load(video_path, sr=None, mono=False)
        sample_interval = int(sample_rate * self._interval)
        if len(waveform.shape) == 1:
            y = np.expand_dims(y, axis=0)
        for i, n in enumerate(range(0, waveform.shape[1], sample_interval)):
            n_end = n + sample_interval
            audio_path = str(output_path / f"audio_{i:04d}.wav")
            sf.write(audio_path, waveform[:,n:n_end].T, sample_rate)
            audios["data"].append(audio_path)

        return [frames, audios]



from typing import Any, Dict, List, Tuple, Optional

from radient.utils.lazy_import import LazyImport

MilvusClient = LazyImport("pymilvus", attribute="MilvusClient", min_version="2.4.2")
pymilvus = LazyImport("pymilvus", min_version="2.4.2")  # Milvus Python SDK


class _MilvusInterface(object):
    """Interface to the Milvus vector database.

    This interface also works with Zilliz Cloud (https://zilliz.com/cloud).
    """

    _clients = {}
    _collection_fields = {}

    def __new__(cls, *args, **kwargs):
        return cls._get_client(*args, **kwargs)

    @classmethod
    def _get_client(
        cls,
        milvus_uri: str,
        collection_name: str,
        dimension: Optional[int] = None
    ) -> Tuple["MilvusClient", Dict[str, str]]:

        milvus_uri = milvus_uri.replace("localhost", "127.0.0.1")

        # If a local Milvus installation was specified, check to see if it's up
        # and running first. If not, prompt the user and start an embedded
        # Milvus instance.
        if milvus_uri not in cls._clients:
            pymilvus.connections.connect(uri=milvus_uri)
            cls._clients[milvus_uri] = MilvusClient(uri=milvus_uri)
        client = cls._clients[milvus_uri]

        # Grab the collection information. If it doesn't exist yet, create it
        # with some default settings. With the collection information, we then
        # store the vector field names inside the `_collection_fields` global
        # object.
        uri_and_coll = (milvus_uri, collection_name)
        if uri_and_coll not in cls._collection_fields:
            if not client.has_collection(collection_name=collection_name):
                client.create_collection(
                    collection_name=collection_name,
                    dimension=dimension,
                    auto_id=True,
                    enable_dynamic_field=True
                )
            info = client.describe_collection(collection_name=collection_name)
            fields = {}
            # TODO(fzliu): support multiple vector fields of the same type.
            for f in info["fields"]:
                if f["type"] == pymilvus.DataType.BINARY_VECTOR:
                    fields["binary"] = f["name"]
                elif f["type"] == pymilvus.DataType.FLOAT_VECTOR:
                    fields["dense"] = f["name"]
                elif (pymilvus.__version__ >= "2.4.0" and
                      f["type"] == pymilvus.DataType.SPARSE_FLOAT_VECTOR):
                    fields["sparse"] = f["name"]
            cls._collection_fields[uri_and_coll] = fields
        info = cls._collection_fields[uri_and_coll]

        return (client, info)



from abc import abstractmethod
from typing import List, Union

from radient.tasks._base import Task
from radient.vector import Vector


class Sink(Task):
    """Sinks in Radient are destinations for vector data. The penultimate
    operation prior to sinks is usually the result of some data merging
    function and, in some cases, can be the direct output of a vectorizer or
    set of vectorizers.
    """

    @abstractmethod
    def __init__(self):
        super().__init__()

    def __call__(self, *args, **kwargs):
        return self.transact(*args, **kwargs)

    @abstractmethod
    def transact(
        self,
        vectors: Union[Vector, List[Vector]],
        **kwargs
    ) -> bool:
        pass




__all__ = [
    "TextVectorizer"
]

from abc import abstractmethod
from typing import Any, List

from radient.tasks.vectorizers._base import Vectorizer
from radient.utils import fully_qualified_name
from radient.vector import Vector


class TextVectorizer(Vectorizer):

    @abstractmethod
    def __init__(self):
        super().__init__()

    def _preprocess(self, text: Any, **kwargs) -> str:
        if not isinstance(text, str):
            return str(text)
        return text



from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np

from radient.utils import fully_qualified_name
from radient.utils.lazy_import import LazyImport
from radient.tasks.vectorizers._base import Vectorizer

torch = LazyImport("torch")
onnx = LazyImport("onnx")
ort = LazyImport("onnxruntime", package_name="onnxruntime-gpu")


def export_to_onnx(
    vectorizer: Vectorizer,
    model_args: Union[Tuple[Any, ...], Any],
    axes_names: Sequence[str] = [],
    input_names: Sequence[str] = [],
    output_names: Sequence[str] = [],
    model_type: Optional[str] = None
):
    """Attempts to export a model in ONNX format for use with `onnxruntime`.
    Switches export implementation based on torch, tensorflow, or scikit-learn
    models.
    """

    # If a model type/library was not specified, attempt to programmatically
    # determine it using the object's fully qualified name. This doesn't work
    # for child classes (e.g. inheriting `nn.Module` or `nn.Sequential`) yet.
    if not model_type:
        model_qualified_name = fully_qualified_name(vectorizer.model)
        if "torch.nn" in model_type:
            model_type = "pytorch"
        elif "tensorflow" in model_type:
            model_type = "tensorflow"
        elif "sklearn.feature_extraction" in model_type:
            model_type = "sklearn"
        else:
            raise NotImplementedError

    # Model path example:
    # "~/.radient/accelerated_models/<method_name>/<model_name>.onnx"
    onnx_model_path = Path.home() / ".radient" / "accelerated_models"
    onnx_model_path /= vectorizer.vtype
    onnx_model_path /= vectorizer.model_name + ".onnx"
    onnx_model_path.parent.mkdir(parents=True, exist_ok=True)
    onnx_model_path = str(onnx_model_path)

    if model_type in ("pytorch", "torch"):
        # Generate dynamic axes on-the-fly.
        dynamic_axes = {}
        if input_names and output_names:
            #symbolic_names = {0: "batch_size", 1: "max_seq_len"}
            symbolic_names = dict(zip(range(len(axes_names)), axes_names))
            dynamic_axes.update({k: symbolic_names for k in input_names})
            dynamic_axes.update({k: symbolic_names for k in output_names})
        torch.onnx.export(
            vectorizer.model,
            model_args,
            onnx_model_path,
            do_constant_folding=True,
            input_names=input_names,
            output_names=output_names,
            dynamic_axes=dynamic_axes
        )
    elif model_type in ("tensorflow", "tf"):
        raise NotImplementedError
    elif model_type in ("scikit-learn", "sklearn"):
        raise NotImplementedError
    else:
        raise NotImplementedError

    return onnx_model_path


class ONNXForward(object):
    """Callable object that runs forward inference on an ONNX model.
    """

    def __init__(
        self,
        model_path: str,
        output_names: Optional[List[str]] = None,
        output_class: Optional[Callable] = None,
        providers: Optional[List[str]] = None
    ):
        super().__init__()
        self._session = ort.InferenceSession(model_path, providers=providers)
        self._output_names = output_names
        self._output_class = output_class

    def __call__(
        self,
        inputs: Union[Dict, Sequence, np.ndarray]
    ) -> List[Union[Dict, np.ndarray]]:
        inputs_ = {}
        input_names = [node.name for node in self._session.get_inputs()]
        if isinstance(inputs, dict):
            # For dictionary inputs, ONNX has a tendency to append a `.N` for
            # tensors that have the identical names in Pytorch model
            # definitions. For example:
            #
            # `attention_mask` -> `attention_mask.3`
            #
            # We automatically detect and compensate for these changes here.
            for name, feat in inputs.items():
                is_match = lambda x: name == x.split(".")[0]
                nms = [nm for nm in input_names if is_match(nm)]
                assert len(nms) == 1, "found conflicting input names"
                inputs_[nms[0]] = np.array(feat)
        elif isinstance(inputs, list):
            inputs = [np.array(item) for item in inputs]
            inputs_ = dict(zip(input_names, inputs))
        else:
            inputs_ = {input_names[0]: np.array(inputs)}

        # Optionally cast model outputs to the desired type, e.g. torch.Tensor.
        result = self._session.run(self._output_names, inputs_)
        if self._output_class:
            result = [self._output_class(arr) for arr in result]

        if self._output_names:
            # If output names were specified, return the result as a
            # dictionary rather than a list.
            assert len(result) == len(self._output_names), "length mismatch"
            result_ = dict(zip(self._output_names, result))
        elif len(result) == 1:
            result_ = result[0]
        else:
            result_ = result
        return result_



__all__ = [
    "CohereTextVectorizer"
]

import os
from typing import List, Optional

from radient.utils.lazy_import import LazyImport
from radient.vector import Vector
from radient.tasks.vectorizers.text._base import TextVectorizer

cohere = LazyImport("cohere")


class CohereTextVectorizer(TextVectorizer):
    """Text vectorization with Cohere (https://www.cohere.com).
    """

    def __init__(self, model_name: str = "embed-english-v3.0", **kwargs):
        super().__init__()
        self._model_name = model_name
        if "COHERE_API_KEY" in os.environ:
            api_key = os.envirion["COHERE_API_KEY"]
        elif api_key in kwargs:
            api_key = kwargs["api_key"]
        else:
            raise ValueError("API key not found")
        self._client = cohere.Client(api_key)

    def _vectorize(self, text: str, **kwargs) -> Vector:
        vector = self._client.embed([text], model=self._model_name)
        return vector.view(Vector)

    @property
    def model_name(self):
        return self._model_name



__all__ = [
    "TorchaudioAudioVectorizer"
]

from typing import List, Tuple

import numpy as np

from radient.tasks.accelerate import export_to_onnx, ONNXForward
from radient.tasks.vectorizers.audio._base import AudioVectorizer
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector

torchaudio = LazyImport("torchaudio")
torch = LazyImport("torch")


class TorchaudioAudioVectorizer(AudioVectorizer):
    """Computes audio embeddings using `torchaudio`.
    """

    def __init__(
        self,
        model_name: str = "HUBERT_BASE",
        reduce_method: str = "avg_pool",
        **kwargs
    ):
        super().__init__()
        self._model_name = model_name
        self._reduce_method = reduce_method
        bundle = getattr(torchaudio.pipelines, model_name)
        self._sample_rate = bundle.sample_rate
        self._model = bundle.get_model()

    def _vectorize(self, audio: np.ndarray, **kwargs) -> List[Vector]:
        with torch.inference_mode():
            output = self._model.forward(torch.from_numpy(audio))
            features = output[0] if isinstance(output, tuple) else output
            if isinstance(features, torch.Tensor):
                features = features.numpy()

            # Torchaudio vectorizers output a list of features, so we
            # optionally reduce the features to a single 1D vector using
            # the method specified by the function caller.
            if self._reduce_method == "avg_pool":
                output = np.mean(features, axis=(0,1)).view(Vector)
            else:
                output = [v.view(Vector) for v in np.mean(features, axis=0)]

            return output

    def accelerate(self):
        # Torchaudio-based vectorizers take an optional `lengths` parameter,
        # which we ignore here.
        onnx_model_path = export_to_onnx(
            self,
            torch.randn((1, 400)),
            axes_names=["batch_size", "seq_len"],
            input_names=["waveforms"],
            output_names=["features"],
            model_type="pytorch"
        )
        self._model.forward = ONNXForward(
            onnx_model_path
        )

    @property
    def sample_rate(self):
        return self._sample_rate



__all__ = [
    "FastRPGraphVectorizer"
]

from typing import Any, List, Sequence

import numpy as np

from radient.tasks.vectorizers.graph._base import GraphVectorizer
from radient.utils.lazy_import import LazyImport
from radient.vector import Vector

sp = LazyImport("scipy")
SparseRandomProjection = LazyImport("sklearn.random_projection", attribute="SparseRandomProjection", package_name="scikit-learn")


class FastRPGraphVectorizer(GraphVectorizer):
    """Computes node (not graph) embeddings using the FastRP algorithm.
    """

    def __init__(
        self,
        dimension: int = 64,
        weights: Sequence = (0.1, 0.2, 1.0, 3.0),
        beta: float = 0.0
    ):
        super().__init__()
        self._projection = SparseRandomProjection(n_components=dimension)
        self._weights = weights
        self._beta = beta

    def _vectorize(self, graph: "sp.sparse.sparray", **kwargs) -> Vector:
        """Radient-specific implementation of the FastRP vectorization
        algorithm.
        """
        projector = self._projection.fit(graph)
        R = projector.components_.T

        # Compute \mathbf{D} as per Chen et al. equation 0:
        # D = sum(Sip, axis=p) if i = j else 0
        L = graph.sum(axis=1)
        L = 0.5 * L**self._beta / graph.shape[0]
        L = sp.sparse.diags_array(L)
        N_k = graph @ L @ R

        # Compute a weighted combination of the powers of the projections.
        result = self._weights[0] * N_k
        for k in range(1, len(self._weights)):
            N_k = graph @ N_k
            result += self._weights[k] * N_k

        result = result.todense().view(Vector)
        return [item for item in result]



from abc import abstractmethod
from collections.abc import Sequence
from typing import Any, Dict, List, Optional, Union
import warnings

import numpy as np

from radient.tasks import Task
from radient.utils import fully_qualified_name
from radient.vector import Vector


def normalize_vector(vector: Vector, inplace: bool = True):
    if not np.issubdtype(vector.dtype, np.floating):
        warnings.warn("non-float vectors are not normalized")
    else:
        if inplace:
            vector /= np.linalg.norm(vector)
        else:
            vector = vector / np.linalg.norm(vector)
    return vector


class Vectorizer(Task):
    """Base class for all vectorizers. Custom vectorizers shouldn't directly
    inherit this class, but should inherit the appropriate subclass e.g.
    `ImageVectorizer` or `AudioVectorizer`.
    """

    @abstractmethod
    def __init__(self):
        super().__init__()
        self._model_name = None
        self._model = None

    def __call__(self, *args, **kwargs):
        vectors = self.vectorize(*args, **kwargs)
        return {"vectors": vectors}

    @property
    def model_name(self) -> Optional[str]:
        return self._model_name

    @property
    def model(self) -> Optional[Any]:
        return self._model

    @property
    def vtype(self) -> str:
        return fully_qualified_name(self).split(".")[3]

    def _preprocess(self, item: Any, **kwargs) -> Any:
        return item

    @abstractmethod
    def _vectorize(self, data: Any, **kwargs) -> Vector:
        pass

    def _postprocess(self, vector: Vector, normalize: bool = True, **kwargs) -> Vector:
        if normalize:
            # Some vectorizers return a _sequence_ of vectors for a single
            # piece of data (most commonly data that varies with time such as
            # videos and audio). Automatically check for these here and
            # normalize them if this is the case.
            if isinstance(vector, Sequence):
                for v in vector:
                    normalize_vector(v)
            else:
                normalize_vector(vector)
        return vector

    def modalities(self) -> List[str]:
        return [self.vtype]

    def vectorize(
        self,
        data: Union[Any, List[Any]],
        modality: Optional[str] = None,
        normalize: bool = True,
        **kwargs
    ) -> Union[Vector, List[Vector], Dict[str, Union[List[Vector], Vector]]]:
        """Vectorizers accept two types of inputs:

        1) One instance of the object/data,
        2) A list of data to be vectorized.

        This function handles both of these cases automatically.
        """
        modality = modality or self.vtype
        if modality in self.modalities():
            data_ = data if isinstance(data, list) else [data]
            vectors = []
            for d in data_:
                v = self._preprocess(d, modality=modality)
                v = self._vectorize(v, modality=modality)
                v = self._postprocess(v, modality=modality, normalize=normalize)
                # TODO(fzliu): only store the original paths, e.g. no base64
                # encodings or long-form text stored as metadata
                v.putmeta("data", str(d)).putmeta("modality", modality)
                vectors.append(v)
            return vectors[0] if not isinstance(data, list) else vectors
        else:
            warnings.warn(f"vectorizer does not support modality: {modality}")

    def accelerate(self):
        warnings.warn("vectorizer does not support acceleration")



from collections import OrderedDict
from collections.abc import Iterator
from itertools import cycle, islice

from typing import Any, Dict, List, Union


def _traverse(data: Union[Any, List[Any]]) -> Iterator:
    """Traverse a nested data structure and yield its elements.
    """
    #if isinstance(data, dict):
    #    for k, v in data.items():
    #        yield from _traverse(v)
    if isinstance(data, list):
        for d in data:
            yield from _traverse(d)
    else:
        yield data


def _datalen(data: Union[Any, List[Any]]) -> int:
    """Returns the length of the input data when used with `_traverse`.
    """
    #if isinstance(data, dict):
    #    return sum([_datalen(v) for v in data.values()])
    if isinstance(data, list):
        return len(data)
    else:
        return 1 if data else 0


def flattened(*args, **kwargs) -> Iterator:
    """For use when `flatten_inputs` evaluates to True. Parses out `dict` and
    `list` inputs so that each individual element is passed as an argument into
    the downstream function. 
    """

    kwargs = OrderedDict(kwargs)

    # Combine `args` and `kwargs` datas into a single list, then cycle through
    # all of them until the maximum length is reached.
    datas = list(kwargs.values()) + list(args)
    maxlen = max([_datalen(d) for d in datas])
    generator = zip(*[islice(cycle(_traverse(d)), maxlen) for d in datas])

    # Recombine the flattened inputs into the original form.
    for inputs in generator:
        flat_kwargs = dict(zip(kwargs.keys(), *inputs))
        flat_args = inputs[len(kwargs):]
        yield flat_args, flat_kwargs


__all__ = [
    "RDKitMoleculeVectorizer"
    "molecule_vectorizer"
]

from typing import Optional

from radient.tasks.vectorizers.molecule._base import MoleculeVectorizer
from radient.tasks.vectorizers.molecule.rdkit import RDKitMoleculeVectorizer


def molecule_vectorizer(method: str = "rdkit", **kwargs) -> MoleculeVectorizer:
    """Creates a text vectorizer specified by `method`.
    """

    # Return a reasonable default vectorizer in the event that the user does
    # not specify one.
    if method in ("rdkit",):
        return RDKitMoleculeVectorizer(**kwargs)
    else:
        raise NotImplementedError



from abc import ABC, abstractmethod
from typing import Any, List


class Task(ABC):
    """Tasks are operators that can include transforms, vectorizers, and sinks.
    Data sources will be supported soon^{TM}.
    """

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def __call__(self, *args, **kwargs) -> List[Any]:
        pass